// Copyright 2008 Dolphin Emulator Project
// Licensed under GPLv2+
// Refer to the license.txt file included.

#pragma once

#include <cstring>
#include <list>
#include <map>
#include <string>
#include <vector>

#include "CommonTypes.h"
#include "StringUtil.h"

struct CaseInsensitiveStringCompare
{
    bool operator() (const std::string& a, const std::string& b) const
    {
        return strcasecmp(a.c_str(), b.c_str()) < 0;
    }
};

class IniFile
{
public:
    class Section
    {
        friend class IniFile;

    public:
        Section() {}
        Section(const std::string& _name) : name(_name) {}

        bool Exists(const std::string& key) const;
        bool Delete(const std::string& key);

        void Set(const std::string& key, const std::string& newValue);
        void Set(const std::string& key, const std::string& newValue, const std::string& defaultValue);

        bool Get(const std::string& key, std::string* value, const std::string& defaultValue = NULL_STRING) const;

        void Set(const std::string& keyName, const char* newValue)
        {
            Set(keyName, std::string(newValue));
        }

        void Set(const std::string& keyName, const bool& newValue)
        {
            Set(keyName, StringFromBool(newValue));
        }

        template<typename T>
        void Set(const std::string& key, const T& newValue)
        {
            Set(key, std::to_string(newValue));
        }

        template<typename T>
        void Set(const std::string& key, const T& newValue, const T& defaultValue)
        {
            Set(key, std::to_string(newValue), defaultValue);
        }

        void Set(const std::string& key, const std::vector<std::string>& newValues);

        template<typename T>
        bool Get(const std::string& key, T* value, const T& defaultValue = T()) const
        {
            std::string temp;
            bool retval = Get(key, &temp);
            if (retval && TryParse(temp, value))
                return true;
            *value = defaultValue;
            return false;
        }

        bool Get(const std::string& key, std::vector<std::string>* values);

        bool operator < (const Section& other) const
        {
            return name < other.name;
        }

    protected:
        std::string name;

        std::vector<std::string> keys_order;
        std::map<std::string, std::string, CaseInsensitiveStringCompare> values;

        std::vector<std::string> lines;
    };

    /**
     * Loads sections and keys.
     * @param filename filename of the ini file which should be loaded
     * @param keep_current_data If true, "extends" the currently loaded list of sections and keys with the loaded data (and replaces existing entries). If false, existing data will be erased.
     * @warning Using any other operations than "Get*" and "Exists" is untested and will behave unexpectedly
     * @todo This really is just a hack to support having two levels of gameinis (defaults and user-specified) and should eventually be replaced with a less stupid system.
     */
    bool Load(const std::string& filename, bool keep_current_data = false);

    bool Save(const std::string& filename) const;
    bool Save() const
    {
        if (filename.empty())
            return false;
        else
            return Save(filename);
    }

    // Returns true if key exists in section
    bool Exists(const std::string& sectionName, const std::string& key) const;

    template<typename T>
    bool Get(const std::string& sectionName, const std::string& key, T* value, const T& defaultValue = T())
    {
        return GetOrCreateSection(sectionName)->Get(key, value, defaultValue);
    }

    void Set(const std::string& sectionName, const std::string& key, const std::string& newValue, const std::string& defaultValue)
    {
        GetOrCreateSection(sectionName)->Set(key, newValue, defaultValue);
    }

    void Set(const std::string& sectionName, const std::string& key, const char* newValue)
    {
        GetOrCreateSection(sectionName)->Set(key, std::string(newValue));
    }

    void Set(const std::string& sectionName, const std::string& key, const bool& newValue)
    {
        GetOrCreateSection(sectionName)->Set(key, newValue);
    }

    template<typename T>
    void Set(const std::string& sectionName, const std::string& key, const T& newValue)
    {
        GetOrCreateSection(sectionName)->Set(key, std::to_string(newValue));
    }

    template<typename T>
    void Set(const std::string& sectionName, const std::string& key, const T& newValue, const T& defaultValue)
    {
        GetOrCreateSection(sectionName)->Set(key, std::to_string(newValue), defaultValue);
    }

    template<typename T>
    bool GetIfExists(const std::string& sectionName, const std::string& key, T* value)
    {
        if (Exists(sectionName, key))
            return GetOrCreateSection(sectionName)->Get(key, value);

        return false;
    }

    bool GetKeys(const std::string& sectionName, std::vector<std::string>* keys) const;

    void SetLines(const std::string& sectionName, const std::vector<std::string>& lines);
    bool GetLines(const std::string& sectionName, std::vector<std::string>* lines, const bool remove_comments = true) const;

    bool DeleteKey(const std::string& sectionName, const std::string& key);
    bool DeleteSection(const std::string& sectionName);

    void SortSections();
    void Clear() { sections.clear(); }

    const Section* GetSection(const std::string& section) const;
    Section* GetSection(const std::string& section);
    Section* GetOrCreateSection(const std::string& section);

    // This function is related to parsing data from lines of INI files
    // It's used outside of IniFile, which is why it is exposed publicly
    // In particular it is used in PostProcessing for its configuration
    static void ParseLine(const std::string& line, std::string* keyOut, std::string* valueOut);

private:
    std::list<Section> sections;

    std::string* GetLine(const std::string& section, const std::string& key);
    void CreateSection(const std::string& section);
    std::string filename;

    static const std::string& NULL_STRING;
};
